import math
from adap_gl import AdapGL

def normal_0_to_z( z ):
    """
	Kerridge and Cook(1976, Biometrika, 63, 401-403)'s Algorithm

    """    
    p = None
    if z < 0.0:
        raise Exception("Negative z value in normal_0_to_z")
    else:
        if z == 0.0:
            p = 0.0
        else:
            if z > 6.8:
                zz = z * z;
                d = zz + 3 - 1 / (0.22 * zz + 0.704);
                m = 1 - 1 / d;
                p = 0.5 - math.exp(-0.5 * zz) * m / (z * ((2 * math.pi) ** 0.5));
            else:
                p = 0.0
                zz4 = 0.25 * z * z;
                n2 = 2.0;               # initial value of 2n          ; n = 1
                theta2 = 1.0;	        # initial value of theta(2n-2)
                theta1 = zz4;
                p = 1.0;		# sum for n = 0
                p_prev = 0.0;	        # initial value of sub-sum
                while True:
                    theta2 = zz4 * (theta1 - theta2) / n2;
                    n2 += 1.0;
                    p += theta2 / n2;
                    if (p == p_prev):
                        break;
                    theta1 = zz4 * (theta2 - theta1) / n2;
                    n2 += 1.0;
                    p_prev = p;
                    
                p = z * math.exp(-0.5 * zz4) * p / ((2.0 * math.pi) ** 0.5);

    return p

def Cum_NormalKC( z ):
    """
            Cumulative normal distribution with mean 0 and sd 1
    """
    v = None
    if (z > 0.0):
        v = 0.5 + normal_0_to_z(z);
        if (v > 1.0):
            v = 1.0
    elif z == 0.0:
        v = 0.5
    else:
        v = 0.5 - normal_0_to_z(abs(z));
        if (v < 0.0):
            v = 0.0

    return v

def phi( x ):
    """
            The density function of the standard normal distribution
    """
    return math.exp(-0.5 * x * x) / ((2.0 * math.pi) ** 0.5);


"""                 |            |
            Pnp     |     Pcp    |    Ppp
                    |            |
       -------------|------------|------------   
                    |            |
            Pnc     |     Pcc    |    Ppc
                    |            |
       -------------|------------|------------
                    |            |
            Pnn     |     Pcn    |    Ppn
                    |            |
"""


class FuncY_Left:
    """
            Function(y|rho, a)
            x: from -infinity to a 
            
    """
    def __init__(self, rho, a):
        self.r = rho
        self.a = a

    def Fy(self, y):
        vx_left = Cum_NormalKC((self.a - self.r * y) / ((1.0 - (self.r ** 2)) ** 0.5))
        return phi(y) * vx_left

    def Fynp_t(self, t):
        return self.Fy(-math.log(t)) * (1.0/t)

    def Fynn_t(self, t):
        return self.Fy(math.log(t)) * (1.0/t)

class FuncY_Cntr:
    """
            Function(y|rho, a1, a2)
            x: from a1 to a2            

    """
    def __init__(self, rho, a1, a2):
        self.r = rho
        self.a1 = a1
        self.a2 = a2

    def Fy(self, y):
        vx_cntr = Cum_NormalKC((self.a2 - self.r * y) / ((1.0 - (self.r ** 2)) ** 0.5)) \
                    - Cum_NormalKC((self.a1 - self.r * y) / ((1.0 - (self.r ** 2)) ** 0.5))
        return phi(y) * vx_cntr

    def Fycp_t(self, t):
        return self.Fy(-math.log(t)) * (1.0/t)

    def Fycn_t(self, t):
        return self.Fy(math.log(t)) * (1.0/t)

class FuncY_Right:
    """
            Function(y|rho, a)
            x: from a to +infinity 
            
    """
    def __init__(self, rho, a):
        self.r = rho
        self.a = a

    def Fy(self, y):
        vx_left = 1.0 - Cum_NormalKC((self.a - self.r * y) / ((1.0 - (self.r ** 2)) ** 0.5))
        return phi(y) * vx_left

    def Fypp_t(self, t):
        return self.Fy(-math.log(t)) * (1.0/t)

    def Fypn_t(self, t):
        return self.Fy(math.log(t)) * (1.0/t)


accu = 1.0e-10    

def Pnp(a, b, rho):
    Fy = FuncY_Left(rho, a)
    gl = AdapGL()
    v = gl.integral(0.0, math.exp(-b), Fy.Fynp_t, accu)
    return v
        

def Pcp(a1, a2, b, rho):
    Fy = FuncY_Cntr(rho, a1, a2)
    gl = AdapGL()
    v = gl.integral(0.0, math.exp(-b), Fy.Fycp_t, accu)
    return v


def Ppp(a, b, rho):
    Fy = FuncY_Right(rho, a)
    gl = AdapGL()
    v = gl.integral(0.0, math.exp(-b), Fy.Fypp_t, accu)
    return v

def Pnc(a, b1, b2, rho):
    Fy = FuncY_Left(rho, a)
    gl = AdapGL()
    v = gl.integral(b1, b2, Fy.Fy, accu)
    return v

def Pcc(a1, a2, b1, b2, rho):
    Fy = FuncY_Cntr(rho, a1, a2)
    gl = AdapGL()
    v = gl.integral(b1, b2, Fy.Fy, accu)
    return v
    

def Ppc(a, b1, b2, rho):
    Fy = FuncY_Right(rho, a)
    gl = AdapGL()
    v = gl.integral(b1, b2, Fy.Fy, accu)
    return v
    

def Pnn(a, b, rho):
    Fy = FuncY_Left(rho, a)
    gl = AdapGL()
    v = gl.integral(0.0, math.exp(b), Fy.Fynn_t, accu)
    return v

def Pcn(a1, a2, b, rho):
    Fy = FuncY_Cntr(rho, a1, a2)
    gl = AdapGL()
    v = gl.integral(0.0, math.exp(b), Fy.Fycn_t, accu)
    return v


def Ppn(a, b, rho):
    Fy = FuncY_Right(rho, a)
    gl = AdapGL()
    v = gl.integral(0.0, math.exp(b), Fy.Fypn_t, accu)
    return v


